/*
 * xrick/src/sysevt.c
 *
 * Copyright (C) 1998-2002 BigOrno (bigorno@bigorno.net). All rights reserved.
 *
 * The use and distribution terms for this software are contained in the file
 * named README, which can be found in the root of this distribution. By
 * using this software in any fashion, you are agreeing to be bound by the
 * terms of this license.
 *
 * You must not remove this notice, or any other, from this software.
 *
 * 32X changes by Chilly Willy
 */

/*
 * 20021010 SDLK_n replaced by SDLK_Fn because some non-US keyboards
 *          requires that SHIFT be pressed to input numbers.
 */

#include "32x.h"
#include "hw_32x.h"

#include "system.h"
#include "config.h"
#include "game.h"
#include "debug.h"

#include "control.h"
#include "draw.h"

#define SETBIT(x,b) x |= (b)
#define CLRBIT(x,b) x &= ~(b)

static U16 previous = 0;

/*
 * Process events, if any, then return
 */
void
sysevt_poll(void)
{
  U16 buttons = MARS_SYS_COMM8;

  if (buttons == previous)
    return;

  if (buttons & syskbd_up)
    SETBIT(control_status, CONTROL_UP);
  else
    CLRBIT(control_status, CONTROL_UP);
  if (buttons & syskbd_down)
    SETBIT(control_status, CONTROL_DOWN);
  else
    CLRBIT(control_status, CONTROL_DOWN);
  if (buttons & syskbd_left)
    SETBIT(control_status, CONTROL_LEFT);
  else
    CLRBIT(control_status, CONTROL_LEFT);
  if (buttons & syskbd_right)
    SETBIT(control_status, CONTROL_RIGHT);
  else
    CLRBIT(control_status, CONTROL_RIGHT);
  if (buttons & syskbd_fire)
    SETBIT(control_status, CONTROL_FIRE);
  else
    CLRBIT(control_status, CONTROL_FIRE);

  if (buttons & syskbd_pause)
    SETBIT(control_status, CONTROL_PAUSE);
  else
    CLRBIT(control_status, CONTROL_PAUSE);
  if (buttons & syskbd_end)
    SETBIT(control_status, CONTROL_END);
  else
    CLRBIT(control_status, CONTROL_END);

#ifdef ENABLE_SOUND
  if (!(buttons & SEGA_CTRL_MODE))
  {
    if (buttons & SEGA_CTRL_X)
      syssnd_toggleMute();
    if (buttons & SEGA_CTRL_Y)
      syssnd_vol(-1);
    if (buttons & SEGA_CTRL_Z)
      syssnd_vol(+1);
  }
#endif
#ifdef ENABLE_CHEATS
  if (buttons & SEGA_CTRL_MODE)
  {
    if (buttons & SEGA_CTRL_X)
      game_toggleCheat(1);
    if (buttons & SEGA_CTRL_Y)
      game_toggleCheat(2);
    if (buttons & SEGA_CTRL_Z)
      game_toggleCheat(3);
  }
#endif

  previous = buttons;
}

/*
 * Wait for an event, then process it and return
 */
void
sysevt_wait(void)
{
  while (previous == MARS_SYS_COMM8)
    Hw32xDelay(2);

  sysevt_poll();
}

/* eof */
