/*
 * xrick/src/syssnd.c
 *
 * Copyright (C) 1998-2002 BigOrno (bigorno@bigorno.net). All rights reserved.
 *
 * The use and distribution terms for this software are contained in the file
 * named README, which can be found in the root of this distribution. By
 * using this software in any fashion, you are agreeing to be bound by the
 * terms of this license.
 *
 * You must not remove this notice, or any other, from this software.
 *
 * 32X changes by Chilly Willy
 */

#include "32x.h"
#include "hw_32x.h"

#include <stdlib.h>

#ifdef __MSVC__
#include <memory.h>
#endif

#include "config.h"

#ifdef ENABLE_SOUND

#include "system.h"
#include "game.h"
#include "syssnd.h"
#include "debug.h"
#include "data.h"

#define ADJVOL(S) ((S)*ssndVol)

U16 __attribute__((aligned(16))) sndbuf[SYSSND_MIXSAMPLES*2*2]; // two buffers of SYSSND_MIXSAMPLES words of stereo pwm audio

static channel_t __attribute__((aligned(16))) channel[SYSSND_MIXCHANNELS];

static U8 isAudioActive = FALSE;
static U8 sndMute = FALSE;  /* mute flag */
static U16 sndVol = SYSSND_MAXVOL*2;  /* internal volume */
static U16 sndUVol = SYSSND_MAXVOL;  /* user-selected volume */

static void end_channel(U8);

/*
 * Callback -- this is also where all sound mixing is done
 */
void syssnd_callback(U32 buff)
{
  U8 c;
  S16 s;
  U32 i;
  U32 *stream = (U32 *)(buff | 0x20000000);
  U8 sisAudioActive = *(U8 *)((U32)&isAudioActive | 0x20000000);
  U16 ssndVol = *(U16 *)((U32)&sndVol | 0x20000000);
  U8 ssndMute = *(U8 *)((U32)&sndMute | 0x20000000);
  channel_t *schannel = (channel_t *)((U32)channel | 0x20000000);

  if (sisAudioActive) {
  for (i = 0; i < SYSSND_MIXSAMPLES; i++) {
    s = 0;
    for (c = 0; c < SYSSND_MIXCHANNELS; c++) {
      if (schannel[c].loop != 0) {  /* channel is active */
        if (schannel[c].len > 0) {  /* not ending */
          s += ADJVOL(*schannel[c].buf - 0x80);
          schannel[c].buf++;
          schannel[c].len--;
        }
        else {  /* ending */
          if (schannel[c].loop > 0) schannel[c].loop--;
          if (schannel[c].loop) {  /* just loop */
            IFDEBUG_AUDIO2(sys_printf("xrick/audio: channel %d - loop\n", c););
            schannel[c].buf = schannel[c].snd->buf;
            schannel[c].len = schannel[c].snd->len;
            s += ADJVOL(*schannel[c].buf - 0x80);
            schannel[c].buf++;
            schannel[c].len--;
          }
          else {  /* end for real */
            IFDEBUG_AUDIO2(sys_printf("xrick/audio: channel %d - end\n", c););
            end_channel(c);
          }
        }
      }
    }
    if (ssndMute)
      stream[i] = ((U32)516<<16) | (U32)516;
    else {
      s >>= 4;
      if (s > 512) s = 512;
      else if (s < -512) s = -512;
      s += 516;
      stream[i] = ((U32)s<<16) | (U32)s;
    }
  }
  } else {
    for (i = 0; i < SYSSND_MIXSAMPLES; i++)
      stream[i] = ((U32)516<<16) | (U32)516;
  }
}

static void
end_channel(U8 c)
{
  channel[c].loop = 0;
  channel[c].snd = NULL;
}

void
syssnd_init(void)
{
  U8 c;

  if (sysarg_args_vol != 0) {
    sndUVol = sysarg_args_vol;
    sndVol = sndUVol << 1;
  }

  for (c = 0; c < SYSSND_MIXCHANNELS; c++)
    channel[c].loop = 0;  /* deactivate */

  isAudioActive = TRUE;
}

/*
 * Shutdown
 */
void
syssnd_shutdown(void)
{
  isAudioActive = FALSE;
}

/*
 * Toggle mute
 *
 * When muted, sounds are still managed but not sent to the dsp, hence
 * it is possible to un-mute at any time.
 */
void
syssnd_toggleMute(void)
{
  sndMute = !sndMute;
}

void
syssnd_vol(S8 d)
{
  if ((d < 0 && sndUVol > 0) ||
      (d > 0 && sndUVol < SYSSND_MAXVOL)) {
    sndUVol += d;
    sndVol = sndUVol << 1;
  }
}

/*
 * Play a sound
 *
 * loop: number of times the sound should be played, -1 to loop forever
 * returns: channel number, or -1 if none was available
 *
 * NOTE if sound is already playing, simply reset it (i.e. can not have
 * twice the same sound playing -- tends to become noisy when too many
 * bad guys die at the same time).
 */
S8
syssnd_play(sound_t *sound, S8 loop)
{
  S8 c;

  if (!isAudioActive) return -1;
  if (sound == NULL) return -1;

  while (MARS_SYS_COMM6 == 3) ;
  MARS_SYS_COMM6 = 2;

  c = 0;
  CacheClearLine(&channel[0]);
  while ((channel[c].snd != sound || channel[c].loop == 0) &&
	 channel[c].loop != 0 &&
	 c < SYSSND_MIXCHANNELS)
  {
    c++;
    CacheClearLine(&channel[c]);
  }
  if (c == SYSSND_MIXCHANNELS)
    c = -1;

  IFDEBUG_AUDIO(
    if (channel[c].snd == sound && channel[c].loop != 0)
      sys_printf("xrick/sound: already playing %s on channel %d - resetting\n",
		 sound->name, c);
    else if (c >= 0)
      sys_printf("xrick/sound: playing %s on channel %d\n", sound->name, c);
    );

  if (c >= 0) {
    channel[c].loop = loop;
    channel[c].snd = sound;
    channel[c].buf = sound->buf;
    channel[c].len = sound->len;
  }

  MARS_SYS_COMM6 = 1;

  return c;
}

/*
 * Pause
 *
 * pause: TRUE or FALSE
 * clear: TRUE to cleanup all sounds and make sure we start from scratch
 */
void
syssnd_pause(U8 pause, U8 clear)
{
  if (pause == TRUE)
    isAudioActive = FALSE;
  else
    isAudioActive = TRUE;
}

/*
 * Stop a channel
 */
void
syssnd_stopchan(S8 chan)
{
  if (chan < 0 || chan > SYSSND_MIXCHANNELS)
    return;

  while (MARS_SYS_COMM6 == 3) ;
  MARS_SYS_COMM6 = 2;

  CacheClearLine(&channel[chan]);
  if (channel[chan].snd) end_channel(chan);

  MARS_SYS_COMM6 = 1;
}

/*
 * Stop a sound
 */
void
syssnd_stopsound(sound_t *sound)
{
  U8 i;

  if (!sound) return;

  while (MARS_SYS_COMM6 == 3) ;
  MARS_SYS_COMM6 = 2;

  for (i = 0; i < SYSSND_MIXCHANNELS; i++)
  {
    CacheClearLine(&channel[i]);
    if (channel[i].snd == sound) end_channel(i);
  }

  MARS_SYS_COMM6 = 1;
}

/*
 * See if a sound is playing
 */
int
syssnd_isplaying(sound_t *sound)
{
  U8 i, playing;

  while (MARS_SYS_COMM6 == 3) ;
  MARS_SYS_COMM6 = 2;

  playing = 0;
  for (i = 0; i < SYSSND_MIXCHANNELS; i++)
  {
    CacheClearLine(&channel[i]);
    if (channel[i].snd == sound) playing = 1;
  }

  MARS_SYS_COMM6 = 1;

  return playing;
}


/*
 * Stops all channels.
 */
void
syssnd_stopall(void)
{
  U8 i;

  while (MARS_SYS_COMM6 == 3) ;
  MARS_SYS_COMM6 = 2;

  for (i = 0; i < SYSSND_MIXCHANNELS; i++)
  {
    CacheClearLine(&channel[i]);
    if (channel[i].snd) end_channel(i);
  }

  MARS_SYS_COMM6 = 1;
}

/*
 * Load a sound.
 */
void
syssnd_load(sound_t *snd, char *name)
{
  data_file_t *fd;

  memset(snd, 0, sizeof(sound_t));

  fd = data_file_open(name);
  if (fd)
  {
    U8 p[4];
#ifdef DEBUG
    snd->name = malloc(strlen(name) + 1);
    strncpy(snd->name, name, strlen(name) + 1);
#endif
    data_file_seek(fd, 0x28, SEEK_SET);
    data_file_read(fd, p, 1, 4);
    snd->len = (p[3]<<24) | (p[2]<<16) | (p[1]<<8) | p[0];
    snd->buf = data_file_mmap(fd, 0x2C);
    snd->valid = TRUE;
  }
}

/*
 *
 */
void
syssnd_free(sound_t *s)
{
  s->valid = FALSE;
#ifdef DEBUG
  if (s->name)
    free(s->name);
  s->name = NULL;
#endif
}

#endif /* ENABLE_SOUND */

/* eof */

