/*
 * xrick/src/sysvid.c
 *
 * Copyright (C) 1998-2002 BigOrno (bigorno@bigorno.net). All rights reserved.
 *
 * The use and distribution terms for this software are contained in the file
 * named README, which can be found in the root of this distribution. By
 * using this software in any fashion, you are agreeing to be bound by the
 * terms of this license.
 *
 * You must not remove this notice, or any other, from this software.
 *
 * 32X changes by Chilly Willy
 */

#include <stdlib.h> /* malloc */

#include "32x.h"
#include "hw_32x.h"

#include "system.h"
#include "game.h"
#include "img.h"
#include "debug.h"

#ifdef __MSVC__
#include <memory.h> /* memset */
#endif

U8 *sysvid_fb; /* frame buffer */

//rect_t SCREENRECT = {0, 0, SYSVID_WIDTH, SYSVID_HEIGHT, NULL}; /* whole fb */

static U16 palette[256];

/*
 * color tables
 */

#ifdef GFXPC
static U8 RED[] = { 0x00, 0x50, 0xf0, 0xf0, 0x00, 0x50, 0xf0, 0xf0 };
static U8 GREEN[] = { 0x00, 0xf8, 0x50, 0xf8, 0x00, 0xf8, 0x50, 0xf8 };
static U8 BLUE[] = { 0x00, 0x50, 0x50, 0x50, 0x00, 0xf8, 0xf8, 0xf8 };
#endif
#ifdef GFXST
static U8 RED[] = { 0x00, 0xd8, 0xb0, 0xf8,
                    0x20, 0x00, 0x00, 0x20,
                    0x48, 0x48, 0x90, 0xd8,
                    0x48, 0x68, 0x90, 0xb0,
                    /* cheat colors */
                    0x50, 0xe0, 0xc8, 0xf8,
                    0x68, 0x50, 0x50, 0x68,
                    0x80, 0x80, 0xb0, 0xe0,
                    0x80, 0x98, 0xb0, 0xc8
};
static U8 GREEN[] = { 0x00, 0x00, 0x6c, 0x90,
                      0x24, 0x48, 0x6c, 0x48,
                      0x6c, 0x24, 0x48, 0x6c,
                      0x48, 0x6c, 0x90, 0xb4,
		      /* cheat colors */
                      0x54, 0x54, 0x9c, 0xb4,
                      0x6c, 0x84, 0x9c, 0x84,
                      0x9c, 0x6c, 0x84, 0x9c,
                      0x84, 0x9c, 0xb4, 0xcc
};
static U8 BLUE[] = { 0x00, 0x00, 0x68, 0x68,
                     0x20, 0xb0, 0xd8, 0x00,
                     0x20, 0x00, 0x00, 0x00,
                     0x48, 0x68, 0x90, 0xb0,
		     /* cheat colors */
                     0x50, 0x50, 0x98, 0x98,
                     0x68, 0xc8, 0xe0, 0x50,
                     0x68, 0x50, 0x50, 0x50,
                     0x80, 0x98, 0xb0, 0xc8};
#endif


void
sysvid_setPalette(img_color_t *pal, U16 n)
{
  U16 i;
  volatile U16 *hwpal = &MARS_CRAM;

  for (i = 0; i < n; i++)
    palette[i] = ((pal[i].b >> 3) << 10) | ((pal[i].g >> 3) << 5) | (pal[i].r >> 3);

  for (i = 0; i < 255; i++)
    hwpal[i] = palette[i];
}

void
sysvid_restorePalette()
{
  U16 i;
  volatile U16 *hwpal = &MARS_CRAM;

  for (i = 0; i < 255; i++)
    hwpal[i] = palette[i];
}

void
sysvid_setGamePalette()
{
  U8 i;
  img_color_t pal[256];

  for (i = 0; i < 32; ++i) {
    pal[i].r = RED[i];
    pal[i].g = GREEN[i];
    pal[i].b = BLUE[i];
  }
  sysvid_setPalette(pal, 32);
}

/*
 * Initialise video
 */
void
sysvid_init(void)
{
  IFDEBUG_VIDEO(printf("xrick/video: start\n"););

  Hw32xInit(MARS_VDP_MODE_256, 0);

  sysvid_fb = malloc(SYSVID_WIDTH*SYSVID_HEIGHT);

  IFDEBUG_VIDEO(printf("xrick/video: ready\n"););
}

/*
 * Shutdown video
 */
void
sysvid_shutdown(void)
{
  sysvid_fb = NULL;
}

/*
 * Update screen
 * NOTE errors processing ?
 */
void
sysvid_update(rect_t *rects)
{
  static U8 flipped = 0;

//  if (rects == NULL)
//    return;

  if (flipped)
    Hw32xFlipWait();

  fast_memcpy((void *)(0x24000200 + 320*12), sysvid_fb, SYSVID_WIDTH*SYSVID_HEIGHT/4);

  flipped = 1;
  Hw32xScreenFlip(0);
}


/*
 * Clear screen
 * (077C)
 */
void
sysvid_clear(void)
{
  memset(sysvid_fb, 0, SYSVID_WIDTH * SYSVID_HEIGHT);
}

/* eof */



